from pysyncrosim.session import Session
from pysyncrosim.library import Library
from pysyncrosim.project import Project
from pysyncrosim.scenario import Scenario
from pysyncrosim.raster import Raster
from pysyncrosim.folder import Folder
from pysyncrosim.environment import runtime_input_folder
from pysyncrosim.environment import runtime_output_folder
from pysyncrosim.environment import runtime_temp_folder
from pysyncrosim.environment import progress_bar
from pysyncrosim.helper import library

__version__ = "unknown"
try:  
    from pysyncrosim._version import __version__
except ImportError:
    pass


import pysyncrosim as ps
import numpy as np
import re

class Project(object):
    """
    A class to represent a SyncroSim Project.
        
    """
    __scenarios = None
    __datasheets = None
    
    def __init__(self, pid, name, library):
        """
        Initializes a pysyncrosim Project instance.

        Parameters
        ----------
        pid : Int
            Project ID.
        name : String
            Project name.
        library : Library
            pysyncrosim Library instance.

        Returns
        -------
        None.

        """
        self.__pid = pid
        self.__name = name
        self.__library = library
        self.__owner = None
        self.__date_modified = None
        self.__readonly = None
        self.__info = None
        # All None attributes created by __init_info()
        self.__init_info()
        self.__description = self.__init_description()
        
    @property
    def pid(self):
        """
        Retrieves Project ID.

        Returns
        -------
        Int
            Project ID.

        """
        return self.__pid
    
    @property
    def name(self):
        """
        Retrieves or sets Project name.

        Returns
        -------
        String
            Project name.

        """
        return self.__name 
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--name=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
    
    @property
    def library(self):
        """
        Retrieves the Library the Project belongs to.

        Returns
        -------
        Library
            SyncroSim Library class instance.

        """
        return self.__library
    
    @property
    def info(self):
        """
        Retrieves Project information.

        Returns
        -------
        pandas DataFrame
            Project information.

        """
        return self.__info
    
    @property
    def owner(self):
        """
        Gets or sets the owner of this Project.

        Returns
        -------
        String
            Owner of this Project.

        """
        return self.__owner
    
    @owner.setter
    def owner(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--owner=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
        
    @property
    def date_modified(self):
        """
        Gets the last date this Project was modified.

        Returns
        -------
        String
            Last date modified.

        """
        return self.__date_modified
    
    @property
    def readonly(self):
        """
        Gets or sets the read-only status for this Project.

        Returns
        -------
        String
            "yes" if this Project is read-only, "no" otherwise.

        """
        return self.__readonly
    
    @readonly.setter
    def readonly(self, value):
        if value is True or value == "yes":
            ro = "yes"
        elif value is False or value == "no":
            ro = "no"
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--readonly=%s" % ro, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
    
    @property
    def description(self):
        """
        Gets or sets the Project description.

        Returns
        -------
        String
            Project description.

        """
        return self.__description
    
    @description.setter
    def description(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location,
                "--description=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__description = None
        self.__description = self.__init_description()

    def scenarios(self, name=None, sid=None, optional=False, summary=None,
                  results=False):
        """
        Retrieve a DataFrame of Scenarios in this Project.
        
        Parameters
        ----------
        name : String, optional
            Scenario name. The default is None.
        sid : Int, optional
            Scenario ID. The default is None.
        optional : Logical, optional
            Return optional information. The default is False.
        summary : Logical, optional
            If set to False, then returns all Scenarios as SyncroSim
            Scenario instances. The default is None.

        results : Logical, optional
            Return only a list of Results Scenarios. The default is False.
        
        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Scenario 
            information including Scenario ID, Project ID, Name, and Is Result.
            If `optional=True`, also returns Parent ID, Owner, Last Modified, 
            Read Only, Merge Dependencies, Ignore Dependencies, and Auto Gen
            Tags.

        """
        
        self.__scenarios = self.library.scenarios(name=name, project=self, 
                                                  sid=sid,
                                                  pid=self.pid,
                                                  optional=optional,
                                                  summary=summary,
                                                  results=results)
        
        return self.__scenarios
    
    def folders(self, folder=None, parent_folder=None, create=False):
        """
        Retrieves a dataframe of Folder information for this Project if
        folder=None, otherwise creates or retrieves a SyncroSim Folder 
        object.
        
        Parameters
        ----------
        folder : String or Int, optional
            Folder name or ID. If the Folder name does not currently exist in
            the Project, then a new Folder will be created. The default is None.
        parent_folder : String, Int, or Folder class instance, optional
            Parent Folder name, ID, or Folder object. If provided, then the
            current Folder will be nested within the parent Folder. The
            default is None.
        create : Logical, optional
            If True, creates the Folder if it does not exist. The default is 
            False.
        
        Returns
        -------
        pandas.DataFrame or pysyncrosim.Folder
            Dataframe of Folder information for this Project or
            Folder class instance.

        """
        folder_data = ps.Folder(ssimobject=self, folder=folder, 
                             parent_folder=parent_folder, create=create)
        if folder_data._Folder__folder_id is None:
            return folder_data._Folder__data
        return folder_data

    def datasheets(self, name=None, summary=True, optional=False, empty=False,
                   filter_column=None, filter_value=None, include_key=False,
                   show_full_paths=False, return_hidden=False):
        """
        Retrieves a DataFrame of Project Datasheets.
        
        Parameters
        ----------
        name : String, optional
            Datasheet name. The default is None.
        summary : Logical or String, optional
            Whether to list package Datasheets or core Datasheets. The default
            is True.
        optional : Logical, optional
            Return optional columns. The default is False.
        empty : Logical, optional
            If True, returns an empty Datasheet. The default is False.
        filter_column : String
            The column and value to filter the output Datasheet by 
            (e.g. "TransitionGroupID=20"). The default is None.
        include_key : Logical, optional
            Whether to include the primary key of the Datasheet, corresponding
            to the SQL database. Default is False.
        show_full_paths : Logical, optional
            Whether to show the full path of any external files in the Datasheet.
            Default is False.
        return_hidden : Logical, optional
            If set to True, returns all records in a Datasheet, including those
            hidden from the user. Results in a slower query. Default is False. 

        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Datasheet 
            information including Package, Name, and Display Name.
            If `optional=True`, also returns Scope, Is Single, and Is Output.

        """
        
        self.__datasheets = self.library.datasheets(name, summary, optional,
                                                    empty, "Project",
                                                    filter_column, 
                                                    filter_value, include_key,
                                                    show_full_paths,
                                                    return_hidden, self.pid)
        return self.__datasheets
    
    def delete(self, scenario=None, force=False):
        """
        Deletes a Project or Scenario.

        Parameters
        ----------
        scenario : Scenario, String, or Int, optional
            Scenario to delete. The default is None.
        force : Logical, optional
            If True, does not prompt the user to confirm deletion. The default
            is False.

        Returns
        -------
        None.

        """
        
        self.library.delete(project=self, scenario=scenario, force=force)

    def save_datasheet(self, name, data, append=True, force=False):
        """
        Saves a Project-scoped Datasheet.

        Parameters
        ----------
        name : String
            Name of Datasheet to save.
        data : pandas.DataFrame
            DataFrame to save as Datasheet.
        append : Logical, optional
            If True, appends to existing Datasheet. If False, then the
            user must also specify `force=True` to overwrite the existing
            Datasheet. The default is True.
        force : Logical, optional
            If True, overwrites existing Datasheet. The user should be aware that
            this may also delete other definitions and results, so this argument
            should be used with care. The default is False.

        Returns
        -------
        None.

        """
        
        self.library.save_datasheet(name, data, append, force, "Project", self.pid)
        
    def run(self, scenarios=None, jobs=1, copy_external_inputs=False):
        """
        Runs a list of Scenario objects.

        Parameters
        ----------
        scenarios : Scenario, String, Int, or List
            List of Scenrios, SyncroSim Scenario instance, name of Scenario,
            or Scenario ID.
        jobs : Int, optional
            Number of multiprocessors to use. The default is 1.
        copy_external_inputs : Logical, optional
            If False, then a copy of external input files (e.g. GeoTIFF files)
            is not created for each job. Otherwise, a copy of external inputs 
            is created for each job. Applies only when jobs > 1. The default is
            False.

        Returns
        -------
        result_dict : Dictionary
            Dictionary of Results Scenarios.

        """
        # Type checks
        if scenarios is not None and not isinstance(
                scenarios, ps.Scenario) and not isinstance(
                    scenarios, int) and not isinstance(
                    scenarios, np.int64) and not isinstance(
                        scenarios, str) and not isinstance(
                            scenarios, list):
            raise TypeError(
                "scenarios must be Scenario instance, String, Integer, or List")
        if not isinstance(jobs, int) and not isinstance(jobs, np.int64):
            raise TypeError("jobs must be an Integer")
        
        # Collect output in a dictionary
        result_list = []
        
        if scenarios is None:
                
            # Run all Scenarios in a Project
            scenarios = self.scenarios(summary=False)
            
            if not isinstance(scenarios, list):
                scenarios = [scenarios]

            scn_id_str = self.__create_scenario_id_string(scenarios)
                    
        elif scenarios is not None:
                
            if not isinstance(scenarios, list):
                scenarios = [scenarios]

            scn_id_str = self.__create_scenario_id_string(scenarios)

        args = ["--run", "--lib=%s" % self.library.location, "--jobs=%d" % jobs]
        
        if len(scenarios) > 1:
            args += ["--sids=%s" % scn_id_str]
        else:
            args += ["--sid=%s" % scn_id_str]

        if jobs > 1 and copy_external_inputs is True:
            args += ["--copyextfiles=yes"]

        print(f"Running Scenario(s)")  
        result = self.library.session._Session__call_console(args)

        if result.returncode == 0:
            print("Run successful")

        # Reset Project Scenarios
        self.__scenarios = None

        # Reset and retrieve Scenario results
        for scn in scenarios:
            if isinstance(scn, int) or isinstance(scn, np.int64):
                scn = self.scenarios(scn)
            elif isinstance(scn, str):
                scn = self.scenarios(scn)
            scn._Scenario__results = None
            result_id = scn.results()["ScenarioID"].values[-1]
            result_list.append(self.scenarios(sid=result_id))
            
        if len(result_list) == 1:
            return result_list[0]
        else:                
            return result_list
    
    def copy(self, name=None):
        """
        Creates a copy of an existing Project.

        Parameters
        ----------
        name : String, optional
            Name of the new Project. If no name is given, the copied 
            Project is named after the source Project. The default is None.

        Returns
        -------
        Project
            SyncroSim Project class instance.

        """
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        
        if name is None:
            name = self.name + " - Copy"
        
        args = ["--copy", "--project", "--slib=%s" % self.library.name,
                "--pid=%d" % self.pid, "--name=%s" % name, "--copychildren"]
        
        self.library.session._Session__call_console(args)
        
        # Reset Projects
        self.library._Library__projects = None
        self.library._Library__init_projects()
        p = self.library._Library__get_project(name=name)
        
        return ps.Project(p["ProjectID"].values[0],
                          p["Name"].values[0], self.library)
    
    def create_project_folder(self, folder_name):
        """
        Create a folder within a SyncroSim project
        
        Parameters
        ----------
        folder_name : str
            Name of folder to create
        
        Returns
        -------
        str
            Folder id
        """

        args = ["--create", "--folder", "--lib=%s" % self.library.location,
                "--name=%s" % folder_name, "--tpid=%d" % self.pid]
        out = self.library.session._Session__call_console(args, decode=True)
        folder_id = re.findall(r'\d+', out)[0]

        return folder_id

    def create_nested_folder(self, parent_folder_id, folder_name):
        """
        Create a folder within an existing folder.
        
        Parameters
        ----------
        parent_folder_id : int
            Parent folder id
        folder_name : str
            Name of folder to create
        
        Returns
        -------
        str
            Folder id
        """

        args = ["--create", "--folder", "--lib=%s" % self.library.location,
                "--name=%s" % folder_name, "--tfid=%d" % parent_folder_id]
        out = self.library.session._Session__call_console(args, decode=True)
        folder_id = re.findall(r'\d+', out)[0]

        return int(folder_id)
    
    def __init_info(self):
        # Set projects
        self.library.projects()
        proj_info = self.library._Library__projects
        proj_info = proj_info[proj_info["ProjectID"] == self.pid]
        self.__owner = proj_info["Owner"].item()
        self.__date_modified = proj_info["DateLastModified"].item()
        self.__readonly = proj_info["IsReadOnly"].item()
        self.__info = proj_info.set_axis(
            ["Value"], axis=0
            ).T.rename_axis("Property").reset_index()
        
    def __init_description(self):
        args = ["--list", "--description",
                "--lib=%s" % self.library.location,
                "--pid=%d" % self.pid]
        return self.library.session._Session__call_console(
            args, decode=True)
    
    def __create_scenario_id_string(self, scenarios):
        # Create string list of scenarios to provide SyncroSim
        scn_string_list = ""
        for scn in scenarios:
            if isinstance(scn, ps.Scenario):
                scn_string_list += str(scn.sid) + ","
            elif isinstance(scn, int) or isinstance(scn, np.int64):
                scn_string_list += str(scn) + ","
            elif isinstance(scn, str):
                scn_object = self.scenarios(name=scn)
                scn_string_list += str(scn_object.sid) + ","

        # Remove trailing comma
        scn_string_list = scn_string_list[:-1]

        return scn_string_list





import os
import pandas as pd

def runtime_input_folder(scenario, datasheet_name):
    """
    Creates a SyncroSim Datasheet input folder.

    Parameters
    ----------
    scenario : Scenario
        Scenario class instance.
    datasheet_name : String
        Name of SyncroSim Datasheet.

    Returns
    -------
    String
        Path to input folder.

    """    
    _validate_environment()
    parent_folder = _environment.input_directory.item()
    
    return _create_scenario_folder(scenario, parent_folder, datasheet_name)

def runtime_output_folder(scenario, datasheet_name):
    """
    Creates a SyncroSim Datasheet output folder.

    Parameters
    ----------
    scenario : Scenario
        Scenario class instance.
    datasheet_name : String
        Name of SyncroSim Datasheet.

    Returns
    -------
    String
        Path to ouput folder.

    """
    _validate_environment()
    parent_folder = _environment.output_directory.item()
    
    return _create_scenario_folder(scenario, parent_folder, datasheet_name)

def runtime_temp_folder(folder_name):
    """
    Creates a SyncroSim Datasheet temporary folder.

    Parameters
    ----------
    folder_name : String
        Name of temporary folder.

    Returns
    -------
    String
        Path to temporary folder.

    """
    _validate_environment()
    return _create_temp_folder(folder_name)

def progress_bar(report_type="step", iteration=None, timestep=None,
                 total_steps=None, message=None):
    """
    Begins, steps, ends, and reports progress for a SyncroSim simulation.

    Parameters
    ----------
    report_type : String, optional
        Directive to "begin", "end", "report", "message", or "step" the
        simulation. The default is "step".
    iteration : Int, optional
        Number of iterations. The default is None.
    timestep : Int, optional
        Number of timesteps. The default is None.
    total_steps : Int, optional
        Number of total steps in the simulation. The default is None.
    message : String, optional
        A message to print to the progress bar status. The default is None.

    Raises
    ------
    TypeError
        If iteration, timestep, or total_steps are not Integers.
    ValueError
        If report_type is not "begin", "end", "step", "report", or "message".

    Returns
    -------
    None.

    """
    _validate_environment()
    
    # Begin progress bar tracking
    if report_type == "begin": 
        try:
            assert total_steps % 1 == 0
            total_steps = int(total_steps)
            print("ssim-task-start=%d\r\n" % total_steps, flush=True)
        except AssertionError or TypeError:
            raise TypeError("total_steps must be an Integer")
            
    # End progress bar tracking
    elif report_type == "end":
        print("ssim-task-end=True\r\n", flush=True)
        
    # Step progress bar
    elif report_type == "step":
        print("ssim-task-step=1\r\n", flush=True)
    
    # Report iteration and timestep
    elif report_type == "report":
        try:
            assert iteration % 1 == 0
            assert timestep % 1 == 0
            print(
                f"ssim-task-status=Simulating -> Iteration is {iteration}" +
                " - Timestep is {timestep}\r\n",
                flush=True)
        except AssertionError or TypeError:
            raise TypeError("iteration and timestep must be Integers")

    # Print arbitrary message
    elif report_type == "message":
        print(
            "ssim-task-status=" + str(message) + "\r\n",
            flush=True)
    else:
        raise ValueError("Invalid report_type")

def update_run_log(*message, sep=""):
    """
    Updates the run log for a SyncroSim simulation.

    Parameters
    ----------
    *message : String
        Message to write to the run log. Can be provided as multiple arguments
        that will be concatenated together using sep.
    sep : String, optional
        String to use if concatenating multiple message arguments. The default
        is an empty String.

    Raises
    ------
    ValueError
        If no message is provided.

    Returns
    -------
    None.

    """
    _validate_environment()
    
    # Check that a message is provided
    if len(message) == 0:
        raise ValueError("Please include a message to send to the run log.")
    
    # Concatenate additional message pieces
    if len(message) > 1:
        full_message = ""
        for m in message:
            full_message = full_message + str(sep) + str(m)
    else:
        full_message = message[0]

    # Split the message at line breaks
    split_message = full_message.split("\n")

    # Standardize surrounding empty lines
    if split_message[0] == "":
        split_message = split_message[1:]
    if split_message[-1] != "":
        split_message.append("")

    # Annotate messages
    annotated_message = ["ssim-task-log=" + m + "\r\n" for m in split_message]

    # Send to SyncroSim
    for m in annotated_message:
        print(m, flush=True)

def _environment():
    env_df = pd.DataFrame(
        {"package_directory": [os.getenv("SSIM_PACKAGE_DIRECTORY")],
         "program_directory": [os.getenv("SSIM_PROGRAM_DIRECTORY")],
         "library_filepath": [os.getenv("SSIM_LIBRARY_FILEPATH")],
         "project_id": [int(os.getenv("SSIM_PROJECT_ID", default=-1))],
         "scenario_id": [int(os.getenv("SSIM_SCENARIO_ID", default=-1))],
         "input_directory": [os.getenv("SSIM_INPUT_DIRECTORY")],
         "output_directory": [os.getenv("SSIM_OUTPUT_DIRECTORY")],
         "temp_directory": [os.getenv("SSIM_TEMP_DIRECTORY")],
         "transfer_directory": [os.getenv("SSIM_TRANSFER_DIRECTORY")],
         "before_iteration": [
             int(os.getenv("SSIM_STOCHASTIC_TIME_BEFORE_ITERATION",
                           default=-1))],
         "after_iteration": [
             int(os.getenv("SSIM_STOCHASTIC_TIME_AFTER_ITERATION",
                           default=-1))],
         "before_timestep": [
             int(os.getenv("SSIM_STOCHASTIC_TIME_BEFORE_TIMESTEP",
                           default=-1))],
         "after_timestep": [
             int(os.getenv("SSIM_STOCHASTIC_TIME_AFTER_TIMESTEP",
                           default=-1))]})
    return env_df
    
def _validate_environment():
    e = _environment()
    
    if e.program_directory.item() is None:
        raise RuntimeError("This function requires a SyncroSim environment")
        
def _create_scenario_folder(scenario, parent_folder, datasheet_name):
    sidpart = "Scenario-" + str(scenario.sid)
    # p = re.sub("\\", "/", parent_folder)
    f = os.path.join(parent_folder, sidpart, datasheet_name)
    
    if not os.path.isdir(f):
        os.mkdir(f)
        
    return f

def _create_temp_folder(folder_name):
    t = _environment().temp_directory.item()
    # p = re.sub("\\\\", "/", t)
    f = os.path.join(t, folder_name)
    
    if not os.path.isdir(f):
        os.mkdir(f)
        
    return f




import pysyncrosim as ps
import os
import pandas as pd
import io

def library(name, session=None, package=None, addons=None, template=None,
            forceUpdate=False, overwrite=False, use_conda=None):
    """
    Creates a new SyncroSim Library and opens it as a Library
    class instance.

    Parameters
    ----------
    name : String
        Name of new Library to create.
    session : Session, optional
        Connects the Python session to the SyncroSim executable. If None, then
        creates a Session class instance using the default installation path
        to the SyncroSim executable. The default is None.
    package : String, optional
        The package type. if None selected, then "stsim" will be used.
    addons : String or List of Strings, optional
        One or more addon packages. The default is None.
    template : String, optional
        Creates Library with specified template. The default is None.
    forceUpdate : Logical, optional
        If False, then user is prompted to approve required updates. The 
        default is False.
    overwrite : Logical, optional
        Overwrite existing Library. The default is False.
    use_conda : Logical, optional
        If True, then runs the Library in a conda environment. The 
        default is None, meaning the Library properties determine whether
        the Library is run in a conda environment. When new Libraries are 
        created, the default for use_conda is `False`.

    Returns
    -------
    Library
        SyncroSim Library class instance.

    """
    _validate_library_inputs(name, session, addons, package, template,
                             forceUpdate, overwrite, use_conda)
    
    if session is None:
        session = ps.Session()

    if addons is not None and not isinstance(addons, list):
        addons = [addons]

    name, loc = _configure_library_name(name)

    if os.path.exists(loc) and overwrite is False and package is None:
        _check_library_update(session, loc, forceUpdate)
        return ps.Library(location=loc, session=session)

    # Test that package specified is installed
    if package is None:
        package = "stsim"
        
    installed = session._Session__pkgs
    if package not in installed["Name"].values:
        raise ValueError(f'The package {package} is not installed')
    
    args = ["--create", "--library", "--package=%s" % package,
            "--name=\"%s\"" % loc]
    
    if overwrite is True:      
        args += ["--force"]

    if forceUpdate is True:
        args += ["--update"]

    args += _configure_template_args(session, package, addons, template)
        
    try:
    
        session._Session__call_console(args)
        
        if addons is not None and template is None:
            # Check if addons exists
            for addon in addons:
                if addon not in installed["Name"].values:
                    raise ValueError(
                        f'The addon package {addons} is not installed')
            
                args = ["--create", "--addon", "--lib=%s" % loc,
                        "--name=%s" % addon]

                session._Session__call_console(args)
            
    except ValueError as ve:
        
        print(ve)
        
    except RuntimeError as re1:
        
        re1 = str(re1)
        if "The Library already exists" in re1:
            pass
        else:
            raise RuntimeError(re1)

    _check_library_update(session, loc, forceUpdate)
        
    return ps.Library(location=loc, session=session, use_conda=use_conda)

def _validate_library_inputs(name, session, addons, package, template, 
                             forceUpdate, overwrite, use_conda):
    """
    Validates input types for the create_library function
    """
    if not isinstance(name, str):
        raise TypeError("name must be a String")
    if session is not None and not isinstance(session, ps.Session):
        raise TypeError("session must be None or pysyncrosim Session instance")
    if package is not None and not isinstance(package, str):
        raise TypeError("package must be a String")
    if addons is not None and not isinstance(addons, str):
        if not isinstance(addons, list):
            raise TypeError("addons must be None, a String, or a List")
        if not all(isinstance(addon, str) for addon in addons):
            raise TypeError("addons in list are not all strings")
    if template is not None and not isinstance(template, str):
        raise TypeError("templates must be a String")
    if not isinstance(forceUpdate, bool):
        raise TypeError("forceUpdate must be a Logical")
    if not isinstance(overwrite, bool):
        raise TypeError("overwrite must be a Logical")
    if use_conda is not None and not isinstance(use_conda, bool):
        raise TypeError("use_conda must be None or a Logical")
    
def _configure_library_name(name):
        
    # Add Library extension if not already included
    if name.endswith(".ssim") is False:
        name += ".ssim"
    
    # Check if name is path and if it exists already
    if os.path.split(name)[0] == '':
        loc = os.path.join(os.getcwd(), name)
    elif os.path.isdir(os.path.split(name)[0]):
        loc = name
        name = os.path.split(name)[-1]
    else:
        raise ValueError(f"Path to Library does not exist: {name}")
    
    return name, loc

def _check_library_update(session, loc, forceUpdate):

    try:
        
        args = ["--list", "--addons", "--lib=%s" % loc]
        session._Session__call_console(args)
        return
        
    except RuntimeError as re2:
        
        re2 = str(re2)
        if "The library has unapplied updates" in re2:
            if forceUpdate is False:
                answer = input(f"The Library has unapplied updates. Would you\
                               like to update the library with path {loc}? \
                               (Y/N)")
            else:
                answer = "Y"
                
            if answer == "Y":
                args = ["--update", "--lib=%s" % loc]
                session._Session__call_console(args)
            elif answer == "N":
                raise Exception("Updates not applied and Library not loaded.")
            
def _configure_template_args(session, package, addons, template):
    
    new_args = []

    if template is not None: 
        
        if template.endswith(".ssim") is True:
            template = os.path.splitext(template)[0]
        
        # Check if template exists in base package
        base_temp_args = ["--list", "--templates", "--package=%s" % package]
        base_temps = session._Session__call_console(base_temp_args,
                                                    decode=True,
                                                    csv=True)
        base_temps = pd.read_csv(io.StringIO(base_temps))
        base_temp = package + "_" + template
        
        if addons is not None:
            addon_temp_list = []

            for addon in addons:
                addon_temp_args = ["--list", "--templates", "--package=%s" % addon]

                try:
                    addon_temps = session._Session__call_console(addon_temp_args,
                        decode=True, csv=True)
                except RuntimeError as re:
                    if f"No templates available for package '{addon}'" in re.args[0]:
                        continue

                if len(addon_temp_list) == 0:
                    addon_temps_all = pd.read_csv(io.StringIO(addon_temps))
                else:
                    addon_temps = pd.read_csv(io.StringIO(addon_temps))
                    addon_temps_all = pd.concat([addon_temps_all, addon_temps])

                addon_temp_list.append(addon + "_" + template)
        
        if base_temp in base_temps["Name"].values:
            new_args += ["--template=\"%s\"" % base_temp]
        elif addons is not None:
            for addon_temp in addon_temp_list:
                if addon_temp in addon_temps_all["Name"].values:
                    new_args += ["--template=\"%s\"" % addon_temp]
        else:
            raise ValueError(
                f"Template {template} does not exist in package {package}")
        
    return new_args

def _delete_library(name, session=None, force=False):
    """
    Deletes a SyncroSim Library.

    Parameters
    ----------
    name : String
        Name of SyncroSim Library to delete.
    session : Session, optional
        Connects the Python session to the SyncroSim executable. If None, then
        creates a Session class instance using the default installation path
        to the SyncroSim executable. The default is None.

    Returns
    -------
    None.

    """
    if session is None:
        session = ps.Session()

    if force is False:
        answer = input (f"Are you sure you want to delete {name} (Y/N)?")
    else:
        answer = "Y"
    
    try:
        lib = ps.Library(name, session)
        
        files = [lib._Library__location,
                 lib._Library__location + ".backup",
                 lib._Library__location + ".input",
                 lib._Library__location + ".output",
                 lib._Library__location + ".temp"]
        
        if answer == "Y":
            for f in files:
                if os.path.exists(f):
                    os.remove(f)  
                    
    except (RuntimeError):
        pass
    
def _delete_project(library, name=None, pid=None, session=None,
                    force=False):
    
    if session is None:
        session = ps.Session()
        
    # force statement moved to delete function
    if force is False:
        answer = input (f"Are you sure you want to delete {name} (Y/N)?")
    else:
        answer = "Y"
    
    if answer == "Y":
        
        # Retrieve Project DataFrame
        library._Library__init_projects()
        p = library._Library__get_project(name, pid)
        
        if p.empty:
            if name is not None: 
                raise RuntimeError("The Project does not exist: %s" % name)
            else: 
                raise RuntimeError("The Project does not exist: %d" % pid)
                
        # Delete Project using console   
        if pid is None:
            pid = p["ID"].values[0]
        args = ["--delete", "--project", "--lib=\"%s\"" % library.location,
                "--pid=%d" % pid, "--force"]
        session._Session__call_console(args)
        
        # Reset Projects
        library._Library__projects = None
        library._Library__init_projects()

def _delete_scenario(library, project, name=None, sid=None, session=None,
                     force=False):
    
    # force statement moved to delete function
    if force is False:
        answer = input (f"Are you sure you want to delete {name} (Y/N)?")
    else:
        answer = "Y"
    
    if answer == "Y":
    
        # Retrieve Scenario DataFrame
        library._Library__init_scenarios()
        s = library._Library__get_scenario(name=name, sid=sid)
        
        if s.empty:
            if name is not None: 
                raise RuntimeError("The Scenario does not exist: %s" % name)
            else: 
                raise RuntimeError("The Scenario does not exist: %d" % sid)
                                        
        # Delete Scenario using console   
        if sid is None:
            sid = s["Scenario ID"].values[0]
        args = ["--delete", "--scenario", "--lib=\"%s\"" % library.location,
                "--sid=%d" % sid, "--force"]
        session._Session__call_console(args)
        
        # Reset Scenarios
        library._Library__scenarios = None
        library._Library__init_scenarios()


import pysyncrosim as ps
from pysyncrosim.environment import _environment
import os
import io
import pandas as pd
import numpy as np

class Folder(object):
    """
    A class representing a SyncroSim Folder.
    
    """
    def __init__(self, ssimobject, folder = None, parent_folder = None, create = False):

        self.__validate_inputs(create)

        self.__create = create
        self.__parent_folder = parent_folder

        self.__set_ssimobject_prop(ssimobject)
        self.__data = self.__get_folder_data()
        if self.__project is not None:
            self.__data = self.__data[self.__data["Project ID"] == self.__project.pid]
        if isinstance(folder, str):
            if folder not in self.__data["Name"].values:
                create = True
        self.__set_folder_id_name(folder)

        # Return data if no folder specified
        data_only = False
        if (self.__scenario is None) and (self.__project is None):
            data_only = True
        elif (self.__scenario is None) and (folder is None):
            data_only = True
        
        # Create folder objects if folder provided and ssimobject is project or scenario
        if (self.__name is not None) and (create is True) and (data_only is False):
            self.__create_folder()    
        
    @property
    def folder_id(self):
        """
        Retrieves Folder ID.

        Returns
        -------
        Int
            Folder ID.

        """
        return self.__folder_id
    
    @property 
    def parent_id(self):
        """
        Retrieves Parent Folder ID.

        Returns
        -------
        Int
            Parent ID.
        """
        self.__get_parent_folder_id()

        return self.__parent_id
    
    @property
    def project_id(self):
        """
        Retrieves ID of the Project containing the folder
        
        Returns 
        -------
        Int
            Project ID
        """
        return self.__project.pid
    
    @property
    def name(self):
        """
        Retrieves or sets Folder Name.

        Returns
        -------
        String
            Folder ID.

        """
        info = self.__get_folder_data()
        info_subset = info[info["ID"] == self.folder_id]
        name = info_subset["Name"].values[0]
        return name
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.__library.location, 
                "--name=%s" % value, "--fid=%d" % self.folder_id]
        self.__library.session._Session__call_console(args)

    @property
    def owner(self):
        """
        Retrieves or sets the Folder Owner.
        
        Returns
        -------
        String
            Folder Owner
        """
        info = self.__get_folder_data()
        info_subset = info[info["FolderID"] == self.folder_id]
        owner = info_subset["Owner"].values[0]
        return owner
    
    @owner.setter
    def owner(self, value):
        args = ["--setprop", "--lib=%s" % self.__library.location, 
                "--owner=%s" % value, "--fid=%d" % self.folder_id]
        self.__library.session._Session__call_console(args)

    @property
    def readonly(self):
        """
        Retrieves or sets the read-only status of a Folder.
        
        Returns
        -------
        Bool
            Folder read-only status
        """
        info = self.__get_folder_data()
        info_subset = info[info["ID"] == self.folder_id]
        readonly = info_subset["Read Only"].values[0]
        return readonly
    
    @readonly.setter
    def readonly(self, value):
        if isinstance(value, str):
            value = value.lower()

        if value is True:
            value = "yes"
        elif value is False:
            value = "no"
        elif value == "yes":
            pass
        elif value == "no":
            pass        
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.__library.location, 
                "--readonly=%s" % value, "--fid=%d" % self.folder_id]
        self.__library.session._Session__call_console(args)

    @property
    def description(self):
        """
        Gets or sets the Folder description.

        Returns
        -------
        String
            Folder description.

        """
        info = self.__get_folder_data()
        info_subset = info[info["FolderID"] == self.folder_id]
        description = info_subset["Description"].values[0]
        return description
    
    @description.setter
    def description(self, value):
        args = ["--setprop", "--lib=%s" % self.__library.location,
                "--description=%s" % value, "--fid=%d" % self.folder_id]
        self.__library.session._Session__call_console(args)

    @property
    def date_modified(self):
        """
        Gets the date the Folder was last modified.
        
        Returns
        -------
        String
            Date the Folder was last modified.
        """
        info = self.__get_folder_data()
        info_subset = info[info["FolderID"] == self.folder_id]
        date_modified = info_subset["DateLastModified"].values[0]
        return date_modified

    @property
    def published(self):
        """
        Gets or sets whether the Folder is tagged for publication.

        Returns
        -------
        String
            "yes" if the Folder is tagged for publication and "no" otherwise.
        """
        info = self.__get_folder_data()
        info_subset = info[info["ID"] == self.folder_id]
        published = info_subset["Is Lite"].values[0]
        return published
    
    @published.setter
    def published(self, value):

        if isinstance(value, str):
            value = value.lower()
        if value is True:
            value = "yes"
        elif value is False:
            value = "no"
        elif value == "yes":
            pass
        elif value == "no":
            pass
        else:
            raise TypeError("value must be a Logical")

        args = ["--setprop", "--lib=%s" % self.__library.location,
                "--islite=%s" % value, "--fid=%d" % self.folder_id]
        self.__library.session._Session__call_console(args)

    def __validate_inputs(self, create):
        if not isinstance(create, bool):
            raise ValueError("create must be a boolean value.")

    def __set_ssimobject_prop(self, ssimobject):
        if isinstance(ssimobject, ps.Library):
            self.__scenario = None
            self.__project = None
            self.__library = ssimobject
        elif isinstance(ssimobject, ps.Project):
            self.__scenario = None
            self.__project = ssimobject
            self.__library = ssimobject.library
        elif isinstance(ssimobject, ps.Scenario):
            self.__scenario = ssimobject
            self.__project = ssimobject.project
            self.__library = ssimobject.library

    def __get_folder_data(self):
        args = ["--lib=%s" % self.__library.location, "--list", "--folders"]
        data = self.__library.session._Session__call_console(args, decode=True, csv=True)
        data =  pd.read_csv(io.StringIO(data))
        return data
    
    def __get_parent_folder_id(self):

        lib_structure = self.__library._Library__get_library_structure()   
        f_ind = lib_structure.index[lib_structure['id'] == str(self.folder_id)].tolist()[0]
        f_level = lib_structure.iloc[f_ind]['level']
        parent_id = None
        for i in reversed(range(f_ind)):
            level = lib_structure.iloc[i]['level']
            item = lib_structure.iloc[i]['item']
            if (level < f_level) & (item == "Folder"):
                parent_id = int(lib_structure.iloc[i]['id'])
                break
            elif (level < f_level) & (item == "Project"):
                break          

        self.__parent_id = parent_id
    
    def __set_folder_id_name(self, folder):

        if folder is None:
            self.__name = None
            self.__folder_id = None

        elif isinstance(folder, str):
            self.__name = folder
            data_subset = self.__data[self.__data["Name"] == self.__name]
            if len(data_subset) == 1:
                self.__folder_id = data_subset["ID"].values[0]
            elif (len(data_subset) > 0) & (self.__create is False):
                raise ValueError("Multiple folders with the same name exist. " +
                                 "Use Folder ID instead to retrieve the desired " +
                                 "folder or set create=True to create another folder " + 
                                 "with the same name.")
            
        elif (isinstance(folder, int)) or (isinstance(folder, np.int64)):
            self.__folder_id = folder
            data_subset = self.__data[self.__data["ID"] == self.__folder_id]
            if len(data_subset) == 1:
                self.__name = data_subset["Name"].values[0]
            elif len(data_subset) == 0:
                raise ValueError("Folder ID does not exist.")
            
        else:
            raise ValueError("folder must be a String or Integer.")
        
    def __create_folder(self):
        args = ["--lib=%s" % self.__library.location, "--create", "--folder", 
                "--name=%s" % self.__name]
        
        if self.__parent_folder is not None:
            self.__retrieve_parent_id()
            args += ["--tfid=%s" % str(self.__parent_id)]
        else:
            args += ["--tpid=%s" % str(self.__project.pid)]
            
        out = self.__library.session._Session__call_console(args, decode=True)
        folderId = int(out.split()[3])
        self.__folder_id = folderId

    def __retrieve_parent_id(self):
        if (isinstance(self.__parent_folder, int)) or\
                (isinstance(self.__parent_folder, np.int32)):
            data_subset = self.__data[self.__data["ID"] == self.__parent_folder]
            if len(data_subset) == 1:
                self.__parent_id = self.__parent_folder
                return 
            
        elif isinstance(self.__parent_folder, str):
            data_subset = self.__data[self.data["Name"] == self.__parent_folder]
            if len(data_subset) == 1:
                self.__parent_id = data_subset["Name"].values[0]
                
        elif isinstance(self.__parent_folder, ps.Folder):
            self.__parent_id = self.__parent_folder.folder_id
            return
        
        raise ValueError("parent_folder not found.")   



# Copyright (c) 2024 Apex Resource Management Solution Ltd. (ApexRMS). All rights reserved.
# MIT License

from setuptools import setup, find_packages
from pathlib import Path
from pysyncrosim._version import __version__

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name="pysyncrosim",
      version=__version__,
      description="Python interface to SyncroSim",
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="ApexRMS",
      author_email="info@apexrms.com",
      url="https://syncrosim.com/",
      packages=find_packages(exclude="tests"))

install_requires=['numpy', 'pandas', 'rasterio']

extras_require=['numpy', 'pandas', 'rasterio']


import pandas as pd
import numpy as np
#import sys
import os
import io
import tempfile
import shutil
import pysyncrosim as ps
from pysyncrosim import helper
from pysyncrosim.environment import _environment

pd.set_option("display.max_columns", 50)

class Library(object):
    """
    A class to represent a SyncroSim Library. 
        
    """
    __projects = None
    __scenarios = None
    __datasheets = None
    
    def __init__(self, location=None, session=None, use_conda=None):
        """
        Initializes a pysyncrosim Library instance.

        Parameters
        ----------
        location : String
            Filepath to Library location on disk.
        session : Session
            pysyncrosim Session instance.

        Returns
        -------
        None.

        """
        self.__location = location
        self.__session = session
        self.__use_conda = use_conda

        # Initialize when in a SyncroSim environment
        if location is None and session is None:
            e = _environment()
            if e.program_directory.item() is not None:
                self.__location = e.library_filepath.item()
                self.__session = ps.Session(e.program_directory.item())
            else:
                raise RuntimeError("Not in a SyncroSim environment." +
                                   " Please specify a location and session.")

        if self.__session is None:
            self.__session = ps.Session()

        if self.__use_conda is not None:
            self.__init_conda()

        self.__name = os.path.basename(self.__location)
        self.__addons = self.__init_addons()
        self.__info = None
        self.__package = None
        self.__owner = None
        self.__date_modified = None
        self.__readonly = None
        # All above attributes get created by below function
        self.__init_info()
        self.__description = self.__init_description()
        
        # Initialize projects and scenarios
        self.scenarios()
        self.projects()

    @property
    def session(self):
        """
        Retrieves the Session associated with this Library.

        Returns
        -------
        Session
            The Session for this Library.

        """
        return self.__session
    
    @property
    def name(self):
        """
        Retrieves or sets the name of this Library.

        Returns
        -------
        String
            Library name.

        """
        return self.__name
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.location, "--name=%s" % value]
        self.session._Session__call_console(args)
        self.__name = value
        
    @property
    def location(self):
        """
        Retrieves the file path to this Library.

        Returns
        -------
        String
            Library file path.

        """
        return self.__location

    @property
    def use_conda(self):
        """
        Retrieves the Conda environment for this Library.

        Returns
        -------
        String
            Conda environment name.

        """
        return self.__use_conda

    @use_conda.setter
    def use_conda(self, value):
        self.__use_conda = value
        self.__init_conda()
    
    @property
    def package(self):
        """
        Retrieves the package this Library is using.

        Returns
        -------
        String
            Package name.

        """
        return self.__package
    
    @property
    def addons(self):
        """
        Retrieves the addon(s) this Library is using.

        Returns
        -------
        String
            Addon name(s).

        """
        return self.__addons
    
    @property
    def info(self):
        """
        Retrieves information about this Library.

        Returns
        -------
        pd.DataFrame
            General library information.

        """
        return self.__info
    
    @property
    def owner(self):
        """
        Gets or sets the owner of this Library.

        Returns
        -------
        String
            Owner of this Library.

        """
        return self.__owner
    
    @owner.setter
    def owner(self, value):
        if not isinstance(value, str):
            raise AttributeError("owner must be a String")
        self.__owner = value
        args = ["--setprop", "--lib=%s" % self.location, "--owner=%s" % value]
        self.session._Session__call_console(args)
        self.__owner = value
        self.__init_info()
        
    @property
    def readonly(self):
        """
        Gets or sets the read-only status of this Library.

        Returns
        -------
        String
            "yes" if this Project is read-only, "no" otherwise.

        """
        return self.__readonly
    
    @readonly.setter
    def readonly(self, value):
        if not isinstance(value, bool):
            raise AttributeError("readonly must be a Logical")
        elif value is True or value == "yes":
            self.__readonly = "yes"
        elif value is False or value == "no":
            self.__readonly = "no"
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.location,
                "--readonly=%s" % self.__readonly]
        self.session._Session__call_console(args)
        self.__init_info()
        
    @property
    def description(self):
        """
        Gets or sets the description for this Library.

        Returns
        -------
        String
            Library description.

        """
        return self.__description
    
    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise AttributeError("description must be a String")
        args = ["--setprop", "--lib=%s" % self.location,
                "--description=%s" % value]
        self.session._Session__call_console(args)
        self.__description = self.__init_description()
        
    @property
    def date_modified(self):
        """
        Retrieves the last date this Library was modified.

        Returns
        -------
        String
            Last date modified.

        """
        return self.__date_modified
    
    def projects(self, name=None, pid=None, summary=True, overwrite=False):
        """
        Creates or opens one or more SyncroSim Projects in the Library.

        Parameters
        ----------
        name : String, optional
            Name of Project. The default is None.
        pid : Int, optional
            Project ID. The default is None.
        summary : Logical, optional
            If True, returns a DataFrame of Project information. If False, 
            returns a list of Project objects. Default is True.
        overwrite : Logical, optional
            If True, then overwrites an existing Project. Default is False.

        Returns
        -------
        Project, List of Projects, or pandas.DataFrame
            DataFrame containing Project information, including ID, Name, 
            Owner, Last Modified, and Read Only, or List of Projects if 
            `summary=False`.

        """
        
        self.__validate_projects_inputs(name, pid, summary, overwrite)
        
        if overwrite is True:
            self.delete(project=name)
        
        if self.__projects is None:
            self.__init_projects()
          
        p = self.__get_project(name=name, pid=pid)
        
        if p is None:
            
            if summary is True:
                
                # Return DataFrame of Project information
                project_summary = self.__extract_project_summary(name, pid)
                
                return project_summary
                
            if summary is False:
                
                project_list = self.__extract_projects()
        
                return project_list  
            
        elif p.empty:
            
            new_project = self.__create_new_project(name)
            
            return new_project
                
        else:
            
            existing_project = self.__open_existing_project(p)
             
            return existing_project
                
        
    def scenarios(self, name=None, project=None, sid=None, pid=None,
                  overwrite=False, optional=False, summary=None,
                  results=False):
        """
        Retrieves a Scenario or DataFrame of Scenarios in this Library.

        Parameters
        ----------
        name : String, Int, or List of these, optional
            Scenario name. If an integer is given, the value will be parsed as
            as Scenario ID. The default is None.
        project : Project, String, or Int, optional
            Project the Scenario belongs to. If no Project is specified when
            creating a new Scenario, the "Definitions" default Project is used.
            The default is None.
        sid : Int or List of Ints, optional
            Scenario ID. If both name and sid are specified, the sid is used. 
            The default is None.
        pid : Int, optional
            Project ID. The default is None.
        overwrite : Logical, optional
            Overwrites an existing Scenario. The default is False.
        optional : Logical, optional
            Return optional information. The default is False.
        summary : Logical, optional
            When name and sid is None, if True, returns a DataFrame of 
            information on existing Scenarios. Otherwise returns a list of 
            Scenario class instances.
        results : Logical, optional
            Return only a list of Results Scenarios. The default is False.

        Returns
        -------
        Scenario, List of Scenarios, or pandas.DataFrame
            If a name or sid is specified, returns a Scenario class instance.
            If no name or sid is specified, and summary is set to False,
            returns a list of Scenario class instances. If no name or sid is
            specified and summary is set to True, returns a pandas.DataFrame
            of Scenario information for this Library.

        """
        
        self.__validate_scenarios_inputs(name, sid, project, pid, overwrite,
                                         optional, summary, results)
          
        # Set the summary argument
        if summary is None:
            if name is not None or sid is not None:
                summary = False
            else:
                summary = True
                
        if sid is not None:
            if not isinstance(sid, list):
                sid = [sid]
            if name is not None:
                print("Both name and sid specified - using sid")
                name = None
            output = [self.__extract_scenario(
                name, project, s, pid, overwrite, optional, summary, results
                ) for s in sid]
            
        elif name is not None:
            if not isinstance(name, list):
                name = [name]
            output = [self.__extract_scenario(
                n, project, sid, pid, overwrite, optional, summary, results
                ) for n in name]
            
        else:
            output = self.__extract_scenario(None, project, None, pid,
                                             overwrite, optional, summary,
                                             results)
        if isinstance(output, list) and len(output) == 1:
            return output[0]
        else:
            return output


    def folders(self):
        """
        Retrieves a dataframe of folder information for this Library.

        Returns
        -------
        pandas.DataFrame
            Dataframe of folder information.
        """
        folder_data = ps.Folder(self)
        return folder_data._Folder__data
        
        
    def datasheets(self, name=None, summary=True, optional=False, empty=False,
                   scope="Library", filter_column=None, filter_value=None,
                   include_key=False, show_full_paths=False, return_hidden=False, *ids):
        """
        Retrieves a DataFrame of Library Datasheets.

        Parameters
        ----------
        name : String, optional
            Datasheet name. The default is None.
        summary : Logical, optional
            When set to True return a dataframe of all available package and
            SyncroSim core Datasheets. When set to False returns a list of 
            Datasheet dataframes. The default is True.
        optional : Logical, optional
            Return optional columns. The default is False.
        empty : Logical, optional
            Return an empty Datasheet. The default is False.
        scope : String, optional
            Datasheet scope. Options include "Library", "Project", or 
            "Scenario". The default is "Library".
        filter_column : String
            The column to filter the output Datasheet by. The default is None.
        filter_value : String, Int, or Logical
            The value to filter the filter_column by. The default is None.
        include_key : Logical, optional
            Whether to include the primary key of the Datasheet, corresponding
            to the SQL database. Default is False.
        show_full_paths : Logical, optional
            If set to True, returns the full path of any external files in the 
            Datasheet. Default is False.
        return_hidden : Logical, optional
            If set to True, returns all records in a Datasheet, including those
            hidden from the user. Results in a slower query. Default is False. 

        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Datasheet 
            information including Package, Name, and Display Name.
            If `optional=True`, also returns Scope, Is Single, and Is Output.

        """
        
        self.__validate_datasheets_inputs(name, summary, optional, empty,
                                          filter_column, include_key,
                                          return_hidden)
        
        # TODO: Instead of setting to None here, make sure datasheets are 
        # refreshed when updates are made - saves computational time
        self.__datasheets = None        
        # TODO: Check if datasheet exists in desired scope
        
        args = self.__initialize_export_args(scope, ids, empty, include_key, show_full_paths)
        
        if name is None:
            
            if summary is True:
                
                ds_frame = self.__return_summarized_datasheets(
                    scope, summary, optional, ids)
                
                return ds_frame
        
            if summary is False:
                
                ds_list = self.__return_list_of_full_datasheets(
                    scope, args, return_hidden, ids)
                    
                return ds_list
        
        else:
            
            # If package is not included in name, add it
            name = self.__check_datasheet_name(name)
            
            if filter_column is not None:
            
                filter_column, filter_value = self.__find_filter_column_args(
                    filter_column, filter_value, name, scope, ids)
                
                args += ["--filtercol=%s" % filter_column + "=" + filter_value]
                        
            if return_hidden:
                ds = self.__slow_query_datasheet(name, scope, ids)               
            else:
                ds = self.__fast_query_datasheet(name, scope, args)            
            
            return ds
        
    def delete(self, project=None, scenario=None, force=False):
        """
        Deletes a SyncroSim class instance.

        Parameters
        ----------
        project : Project, String, or Int, optional
            If called from a Library class instance, specify the Project to 
            delete. The default is None.
        scenario : Scenario, String, or Int, optional
            If called from a Scenario class instance, specify the Scenario to
            delete. The default is None.
        force : Logical, optional
            If set to True, does not ask user before deleting SyncroSim class
            instance. The default is False.

        Returns
        -------
        None.

        """
        # after delete, can still access the instance attributes - should
        # create function to set to None?
        # Also, should have method to delete list of Projects or Scenarios?
        
        # type checks
        if project is not None and not isinstance(project, ps.Project):
            if not isinstance(project, int) and not isinstance(
                    project, str) and not isinstance(project, np.int64):
                raise TypeError(
                    "project must be a Project instance, Integer, or String")
        if scenario is not None and not isinstance(scenario, ps.Scenario):
            if not isinstance(scenario, int) and not isinstance(
                    scenario, str) and not isinstance(scenario, np.int64):
                raise TypeError(
                    "scenario must be a Scenario instance, Integer, or String")
        if not isinstance(force, bool):
            raise TypeError("force must be a Logical")
        
        if project is None and scenario is None:
            
            helper._delete_library(name = self.location, session=self.session,
                               force=force)
        
        elif project is not None and scenario is None:
            
            # turn project into project class instance if str or int
            if type(project) is int:
                p = self.projects(pid = project)
            if type(project) is str:
                if project in self.__projects["Name"].values:
                    p = self.projects(name = project)
                else:
                    raise ValueError(f'project {project} does not exist')
            if isinstance(project, ps.Project):
                p = project
            
            helper._delete_project(library=self, name=p.name,
                                   pid=p.pid, session=self.session,
                                   force=force)
            
        elif scenario is not None:
            
            # turn scenario into scenario class instance if str or int
            if type(scenario) is int:
                s = self.scenarios(sid = scenario, project = project)
            if type(scenario) is str:
                if scenario in self.__scenarios["Name"].values:
                    s = self.scenarios(name = scenario, project = project)
                else:
                    raise ValueError(f'scenario {scenario} does not exist')
            if isinstance(scenario, ps.Scenario):
                s = scenario
            
            helper._delete_scenario(library=self, project=s.project, 
                                    name=s.name, sid=s.sid,
                                    session=self.session,
                                    force=force)
    
    def save_datasheet(self, name, data, append=False, force=False, 
                       scope="Library", *ids):
        """
        Saves a pandas DataFrane as a SyncroSim Datasheet.

        Parameters
        ----------
        name : String
            Name of the Datasheet.
        data : pandas DataFrame
            DataFrame of Datasheet values.
        append : Logical, optional
            If set to True, appends the DataFrame to the existing 
            Datasheet (if the Datasheet accepts multiple rows). If False,
            then the user must also specify force as True to overwrite
            the existing Datasheet. Default is False.
        force : Logical, optional
            If set to True while append is False, overwrites the existing
            Datasheet. The user should be aware that this may also delete 
            other definitions and results, so this argument should be used 
            with care. Default is False.
        scope : String, optional
            Scope of the Datasheet. The default is "Library".
        *ids : Int
            If Project- or Scenario-scoped, requires the Project or Scenario
            IDs.

        Returns
        -------
        None.

        """
        # Type checks
        if not isinstance(name, str):
            raise TypeError("name must be a String")
        if not isinstance(data, pd.DataFrame):
            raise TypeError("data must be a pandas DataFrame")
        if not isinstance(append, bool):
            raise TypeError("append must be a Logical")
        if not isinstance(force, bool):
            raise TypeError("force must be a Logical")
        if not isinstance(scope, str):
            raise TypeError("scope must be a String")
          
        # Add package name to Datasheet name if not included
        if name.startswith(self.package) is False:
            if name.startswith("core") is False:
                name = self.package + "_" + name

        # Convert boolean values to "Yes"/"No"
        for col in data:
            if data[col].dtype == bool:
                data[col] = data[col].map({True: "Yes", False: "No"})
            
        # Check if running in a SyncroSim environment from the user interface
        e = _environment()
        transfer_dir = e.transfer_directory.item()
        
        # If running from user interface, save data to transfer directory
        if (transfer_dir is not None) & (append is False):
            fpath = '{}\\SSIM_OVERWRITE-{}.csv'.format(transfer_dir, name)
            data.to_csv(fpath, index=False)
            return
        elif (transfer_dir is not None) & (append is True):
            fpath = '{}\\SSIM_APPEND-{}.csv'.format(transfer_dir, name)
            data.to_csv(fpath, index=False)
            return
        
        # Otherwise export the data to SyncroSim
        else:
            try:
                fpath = None
                delete_or_warn = self.__validate_delete_datasheet(force, append, scope, data)

                if delete_or_warn == "delete":
                    self.__delete_datasheet(scope, name, ids)
                    if data.empty:
                        return
                elif delete_or_warn == "warn":
                    print("WARNING: The force argument must be set to True " 
                          "to overwrite or delete an existing Project or Library "
                          "Datasheet.")
                    return

                fpath = self.__save_datasheet_to_temp(data)

                args = ["--import", "--lib=%s" % self.location,
                        "--sheet=%s" % name, 
                        "--file=%s" % fpath]

                if append:
                    args += ["--append"]
                if scope == "Project":
                    args += ["--pid=%d" % ids]
                if scope == "Scenario":
                    args += ["--sid=%d" % ids]

                result = self.__session._Session__call_console(args)
                
                if result.returncode == 0:
                    print(f"{name} saved successfully")                

            finally:
                if fpath is not None:
                    shutil.rmtree(os.path.dirname(fpath), ignore_errors=True)
            
    def run(self, scenarios=None, project=None, jobs=1,
            copy_external_inputs=False):
        """
        Runs a list of Scenario objects.

        Parameters
        ----------
        scenarios : Scenario, String, Int, or List
            List of Scenrios, SyncroSim Scenario instance, name of Scenario,
            or Scenario ID.
        project : Project, optional
            SyncroSim Project instance, name of Project, or Project ID.
        jobs : Int, optional
            Number of multiprocessors to use. The default is 1.
        copy_external_inputs : Logical, optional
            If False, then a copy of external input files (e.g. GeoTIFF files)
            is not created for each job. Otherwise, a copy of external inputs 
            is created for each job. Applies only when jobs > 1. The default is
            False.

        Returns
        -------
        result_dict : Dictionary
            Dictionary of Results Scenarios.

        """

        self.__validate_run_inputs(scenarios, project, jobs,
                                   copy_external_inputs)
        
        scenario_list = self.__generate_scenarios_list_to_run(scenarios,
                                                              project)

        # Collect output from all runs
        result_list = [scn.run(
            jobs=jobs, copy_external_inputs=copy_external_inputs
            ) for scn in scenario_list]
            
        if len(result_list) == 1:
            return result_list[0]
        else:                
            return result_list
        
    def update(self):
        """
        Updates a SyncroSim Library.

        Returns
        -------
        None.

        """
        try:
            args = ["--update", "--lib=%s" % self.location]
            self.session._Session__call_console(args)
        
        except RuntimeError as e:
            print(e)
        
    def backup(self):
        """
        Creates a backup of a SyncroSim Library.

        Returns
        -------
        None.

        """
        ds = self.datasheets(name = "core_Backup")
        args = ["--lib=%s" % self.location, "--backup"]
        
        if ds.IncludeInput[0] == "Yes":
            args += ["--input"]
            
        if ds.IncludeOutput[0] == "Yes":
            args += ["--output"]
        
        self.session._Session__call_console(args)
        
    def enable_addons(self, name):
        """
        Enable addon package(s) of a SyncroSim Library.

        Parameters
        ----------
        name : String or List
            Name of addon(s).

        Raises
        ------
        ValueError
            addon does not exist in available addons for this Library.

        Returns
        -------
        None.

        """
        
        # Type checks and convert to list
        if not isinstance(name, list):
            if not isinstance(name, str):
                raise TypeError("name must be a String or List of Strings")
            else:
                name = [name]
        elif all(isinstance(x, str) for x in name) is False:
            raise TypeError("all elements in name must be Strings")
            
            
        for a in name:
            args = ["--create", "--addon", "--lib=%s" % self.location,
                    "--name=%s" % a]
            try:
                self.session._Session__call_console(args)
                
            # Convert SyncroSim console error to print output
            except RuntimeError as e:
                print(e)
                return
         
        # Reset addons
        self.__addons = self.__init_addons()
    
    def disable_addons(self, name):
        """
        Disable addon package(s) of a SyncroSim Library.

        Parameters
        ----------
        name : String
            Name of addon package(s).

        Raises
        ------
        ValueError
            addon does not exist in available addons for this Library.

        Returns
        -------
        None.

        """
        
        # Type checks and convert to list
        if not isinstance(name, list):
            if not isinstance(name, str):
                raise TypeError("name must be a String or List of Strings")
            else:
                name = [name]
        elif all(isinstance(x, str) for x in name) is False:
            raise TypeError("all elements in name must be Strings")
            
        for a in name:
            args = ["--delete", "--addon", "--force",
                    "--lib=%s" % self.location, "--name=%s" % a]
            try:
                self.session._Session__call_console(args)
                
            # Convert SyncroSim console error to print output
            except RuntimeError as e:
                print(e)
                return
            
        # Reset addons
        self.__addons = self.__init_addons()

    def __init_conda(self):
        args = ["--setprop", "--lib=%s" % self.location]

        if self.__use_conda is True:
            args += ["--useconda=yes"]
            current_packages = self.__retrieve_lib_packages()
            self.__create_conda_env(current_packages)
        else:
            args += ["--useconda=no"]
        
        self.session._Session__call_console(args)
        
    def __retrieve_lib_packages(self):
        # Retrieves current packages being used by the library
        args = ["--list", "--datasheets", "--lib=%s" % self.location]
        result = self.session._Session__call_console(args, decode=True, csv=True)
        result =  pd.read_csv(io.StringIO(result))
        return np.unique(result["Package"]).tolist()

    def __create_conda_env(self, packages):

        for p in packages:
            args = ["--conda", "--createenv", "--pkg=%s" % p]
            result = self.session._Session__call_console(args)
            result_message = result.stdout.decode('utf-8')
            if (result.returncode != 0) | \
                ("this package does not use Conda environments" in result_message):
                print(result_message)
                self.__use_conda = False
                self.__init_conda()

    def __init_addons(self):   
        # Retrieves addons information
        args = ["--list", "--addons", "--lib=%s" % self.location]
        return self.__console_to_csv(args)
    
    def __init_info(self):
        # Retrieves Library info
        args = ["--list", "--library", "--lib=%s" % self.location]
        lib_info = self.__console_to_csv(args)
        self.__info = lib_info
        
        # Set index and retrieve all properties
        lib_info = lib_info.set_index("Property")
        self.__package = lib_info.loc["Package Name:"].item()
        self.__owner = lib_info.loc["Owner:"].item()
        self.__readonly = lib_info.loc["Read Only:"].item()
        self.__date_modified = lib_info.loc["Last Modified:"].item()
        
    def __init_description(self):  
        # Retrieves the Library description
        args = ["--list", "--description", "--lib=%s" % self.__location]
        return self.session._Session__call_console(args, decode=True)
                
    def __init_projects(self): 
        # Retrieves a list of Projects
        args = ["--list", "--projects", "--lib=%s" % self.__location]
        self.__projects = self.__console_to_csv(args)
            
    def __init_scenarios(self, pid=None):
        # Retrieves a list of Scenarios
        args = ["--list", "--scenarios", "--lib=%s" % self.__location]
        if pid is not None:
            args += ["--pid=%d" % pid]
        self.__scenarios = self.__console_to_csv(args)
            
    def __init_datasheets(self, scope, summary, name=None, args=None):
        # Retrieves a list of Datasheets
        if self.__datasheets is None:
            if summary is True:
                args = ["--list", "--datasheets", "--lib=%s" % self.__location,
                        "--scope=%s" % scope, "--includesys"]
            self.__datasheets = self.__console_to_csv(args)    
            
    def __check_datasheet_name(self, name):
        # Appends package name to Datasheet name
        enabled_addons = self.addons[self.addons.Enabled == "Yes"]
        enabled_addons = enabled_addons["Name"].tolist()

        for addon_name in enabled_addons:
            if name.startswith(addon_name):
                return name
            
        if name.startswith(self.package):
            return name
        
        if name.startswith("core"):
            return name
        
        return self.package + "_" + name
            
    def __get_project(self, name=None, pid=None):
        # Retrieves Project info from the name or ID
        if name is None and pid is None:
            return None
        if name is not None:
            return self.__projects[self.__projects["Name"] == name]
        else:
            return self.__projects[self.__projects["ProjectID"] == pid]
        
    def __refresh_projects(self):
        # Call SyncroSim DB to refresh project list
        self.__projects = None
        self.__init_projects()
            
    def __get_scenario(self, name=None, sid=None):
        # Retrieves Scenario info from the name or ID
        if name is None and sid is None:
            return None
        if name is not None and sid is not None:
            return self.__scenarios[
                (self.__scenarios["Name"] == name) & (self.__scenarios["ScenarioID"] == sid)]
        if name is not None:
            return self.__scenarios[self.__scenarios["Name"] == name]
        else:
            return self.__scenarios[self.__scenarios["ScenarioID"] == sid]
        
    def __refresh_scenarios(self):
        # Call SyncroSim DB to refresh scenario list
        self.__scenarios = None
        self.__init_scenarios()
        
    def __extract_scenario(self, name, project, sid, pid, overwrite, optional,
                           summary, results):
        
        # Find out if first argument is a Scenario ID
        if isinstance(name, int):
            
            if sid is not None:
                raise ValueError("Name is specified as a Scenario ID, but " + 
                                 "sid is already given")
                
            sid = name
            name = None
        
        # Find project if not specified
        if summary is False:
            
            if pid is None:
                
                pid = self.__find_project_id(project)
                
            if overwrite is True:
                if project is None:                
                    self.delete(project=pid, scenario=name, force=True)
                else:
                    self.delete(project=project, scenario=name, force=True)
                    
        # Retrieve Scenario DataFrame
        self.__init_scenarios(pid=pid)
        
        # If sid provided, check that sid exists
        if sid is not None:
            
            self.__validate_sid(sid, name)
        
        s = self.__get_scenario(name=name, sid=sid)
        
        if (s is None) or (summary is True):
            
            # Retrieve DataFrame of available Scenarios
            if summary:
                
                scenario_summary = self.__extract_scenario_summary(optional,
                                                                   results,
                                                                   name, sid)
                
                return scenario_summary
              
            # Return list of Scenario objects    
            if summary is False:
                
                scenario_list = self.__extract_scenario_list(project, pid) # better name needed
                
                return scenario_list                
    
        # Create a Scenario
        if s.empty:
            
            if project is None:
                project = pid
            
            new_scenario = self.__create_new_scenario(project, name, sid)
            
            return new_scenario
        
        # Open a Scenario
        else:
            
            if project is None:
                project = pid
            
            existing_scenario = self.__open_existing_scenario(s, project)
                            
            return existing_scenario

        
    def __console_to_csv(self, args, index_col=None):
        # Turns console output into a pd.DataFrame
        console_output = self.session._Session__call_console(
            args,
            decode=True,
            csv=True)
        
        return pd.read_csv(io.StringIO(console_output), index_col=index_col)
    
    def __validate_pid(self, pid, name):
        
        # If Scenario specified before project, then project should be created
        # if (self.__scenarios is not None and not self.__scenarios.empty):
        #     if pid == self.__scenarios["ProjectID"].item():
        #         return
        
        if self.__projects is None or self.__projects.empty:
            raise ValueError("pid specified, but no Projects created yet")
        elif pid not in self.__projects["ProjectID"].values:
            raise ValueError(f"Project ID {pid} does not exist")
        elif name is not None and name != self.__projects[
                self.__projects["ProjectID"] == pid]["Name"].values[0]:
            raise ValueError(
                f"Project ID {pid} does not match Project name {name}")
            
    def __validate_sid(self, sid, name):
    
        if self.__get_scenario(sid=sid).empty:
            raise ValueError(f"Scenario ID {sid} does not exist")
        elif name is not None and name != self.__get_scenario(
                sid=1).Name.item():
            raise ValueError(
                f"Scenario ID {sid} does not match Scenario name {name}")
            
    def __extract_project_summary(self, name, pid):
    
        if name is not None:
            return self.__projects[self.__projects.Name == name]
        if pid is not None:
            return self.__projects[self.__projects.ProjectID == pid]
        else:
            return self.__projects
        
    def __extract_projects(self):
        
        self.__refresh_projects()

        proj_list = []
        
        for i in range(0, len(self.__projects)):
            
            proj = ps.Project(self.__projects["ProjectID"].loc[i],
                              self.__projects["Name"].loc[i],
                              self)
            
            proj_list.append(proj)
    
        return proj_list
        
    def __create_new_project(self, name):
     
        # If specified Project does not exist, then create it    
        args = ["--create", "--project", "--lib=%s" % self.__location,
                "--name=%s" % name]
        self.session._Session__call_console(args)
        
        # Reset Projects
        self.__refresh_projects()
        p = self.__get_project(name, pid=None)
        
        # Convert np.int64 to native int
        pid = p["ProjectID"].values[0].tolist()
        
        new_project = ps.Project(pid, p["Name"].values[0], self)
        
        return new_project
    
    def __open_existing_project(self, project):
    
        # Convert np.int64 to native int
        pid = project["ProjectID"].values[0].tolist()
         
        return ps.Project(pid, project["Name"].values[0], self)
    
    def __find_project_from_scenario(self, sid, name):
            
        if sid is not None and self.__get_scenario(sid=sid).empty is False:
            pid = self.__get_scenario(sid=sid)["ProjectID"].item()
            project = self.projects(pid=pid)
        elif name is not None and len(self.__get_scenario(name=name)) == 1:
            pid = self.__get_scenario(name=name)["ProjectID"].item()
            project = self.projects(pid=pid)
        elif self.__projects is None or self.__projects.empty:
            project = self.projects(name = "Definitions")
            pid = project.pid
        elif len(self.__projects) == 1:
            pid = self.__projects.ProjectID.item()
            project = self.projects(pid=pid)     
        else:
            raise ValueError("More than one Project in Library." + 
                             " Please specify a Project.")
            
        return project
    
    def __find_project_id(self, project):
    
        if isinstance(project, int) or isinstance(project, np.int64):
            pid = project
        elif isinstance(project, str):
            # may need to refresh project here
            pid = self.__projects[
                self.__projects.Name == project].ProjectID.item()
        elif isinstance(project, ps.Project):
            pid = project.pid
        elif project is None:
            self.projects()
            if len(self.__projects) > 1:
                raise ValueError("More than one Project in Library." +
                                 " Please specify a Project.")
            elif len(self.__projects) == 0:
                self.__create_new_project("Definitions")
            pid = self.__projects["ProjectID"].item()
            
        return pid
    
    def __extract_scenario_summary(self, optional, results, name, sid):
    
        if optional is False:
            ds =  self.__scenarios[['ScenarioID',
                                    'ProjectID',
                                    'Name',
                                    'IsResult']]
        else:
            ds = self.__scenarios
            
        if results:
            ds = ds[ds["IsResult"] == "Yes"]
        
        if name is not None:
            return ds[ds.Name == name]
        
        if sid is not None:
            return ds[ds["ScenarioID"] == sid]
        
        return ds
    
    def __extract_scenario_list(self, project, pid): # better name needed
    
        s_summary = self.__scenarios
        s_summary = s_summary[s_summary["IsResult"] == "No"]
        
        if not isinstance(project, ps.Project):
            project = self.projects(pid=pid) 
                        
        scn_list = []
        
        for i in range(0, len(s_summary)):
            
            scn = ps.Scenario(s_summary["ScenarioID"].loc[i],
                              s_summary["Name"].loc[i],
                              project, self)
            
            scn_list.append(scn)
    
        return scn_list
    
    def __create_new_scenario(self, project, name, sid):
    
        pid = self.__find_project_id(project)
         
        # Create Scenario using console    
        args = ["--create", "--scenario", "--pid=%d" % pid,
                "--lib=%s" % self.__location, "--name=%s" % name]
        self.session._Session__call_console(args)
        
        # Reset Scenarios
        self.__refresh_scenarios()
        s = self.__get_scenario(name=name, sid=sid)
        
        if not isinstance(project, ps.Project):
            project = self.projects(pid=pid)
        
        return ps.Scenario(s["ScenarioID"].values[0],
                           s["Name"].values[0], project, self)
    
    def __open_existing_scenario(self, scenarios, project):
                    
        # Retrieve the name of a Scenario using only the sid
        pid = scenarios["ProjectID"].values[0].tolist()
    
        if not isinstance(project, ps.Project):
            project = self.projects(pid=pid)
    
        sid = scenarios["ScenarioID"].values[0].tolist()
        
        return ps.Scenario(sid, scenarios["Name"].values[0], project, self)
    
    def __validate_projects_inputs(self, name, pid, summary, overwrite):
        
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        if pid is not None and not isinstance(
                pid, int) and not isinstance(pid, np.int64):
            raise TypeError("pid must be an Integer")
        if not isinstance(summary, bool):
            raise TypeError("summary must be a Logical")
        if not isinstance(overwrite, bool):
            raise TypeError("overwrite must be a Logical")
            
        if pid is not None:
            self.__validate_pid(pid, name)
    
    def __validate_scenarios_inputs(self, name, sid, project, pid, overwrite,
                                    optional, summary, results):

        if name is not None and not isinstance(name, str)\
            and not isinstance(name, int)\
                and not isinstance(name, np.int64)\
                    and not isinstance(name, list):
            raise TypeError("name must be a String, Integer, or List of these")
            
        if isinstance(name, list) and isinstance(sid, list):
            name = None
            
        if isinstance(name, list):
           if not all(isinstance(item, int) for item in name)\
               and not all(isinstance(item, str) for item in name):
               raise TypeError("All values in name list must be either" + 
                               " Strings or Integers")
               
        if project is not None and project.__class__.__name__ != "Project":
            if not isinstance(project, str) and not isinstance(project, int):
                raise TypeError(
                    "project must be Project instance, String, or Integer")
                
        if sid is not None and not isinstance(sid, int)\
            and not isinstance(sid, np.int64)\
                and not isinstance(sid, list):
            raise TypeError("sid must be an Integer or List of Integers")
            
        if isinstance(sid, list):
            if not all(isinstance(item, int) for item in sid):
                raise TypeError("All values in sid list must be Integers")
                
        if pid is not None and not isinstance(
                pid, int) and not isinstance(pid, np.int64):
            raise TypeError("pid must be an Integer")
            
        if not isinstance(overwrite, bool):
            raise TypeError("overwrite must be a Logical")
            
        if not isinstance(optional, bool):
            raise TypeError("optional must be a Logical")
            
        if not isinstance(summary, bool) and summary is not None:
            raise TypeError("summary must be a Logical or None")
            
        if not isinstance(results, bool):
            raise TypeError("results must be a Logical")
    
    def __validate_datasheets_inputs(self, name, summary, optional, empty,
                                     filter_column, include_key, 
                                     return_hidden):
            
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        if not isinstance(summary, bool):
            raise TypeError("summary must be a Logical")
        if not isinstance(optional, bool):
            raise TypeError("optional must be a Logical")
        if not isinstance(empty, bool):
            raise TypeError("empty must be a Logical")
        if not isinstance(filter_column, str) and filter_column is not None:
            raise TypeError("filter_column must be a String")
        if not isinstance(include_key, bool):
            raise TypeError("include_key must be a Logical")
        if not isinstance(return_hidden, bool):
            raise TypeError("return_hidden must be a Logical")
            
    def __validate_run_inputs(self, scenarios, project, jobs,
                               copy_external_inputs):
    
        if scenarios is not None and not isinstance(
                scenarios, ps.Scenario) and not isinstance(
                    scenarios, int) and not isinstance(
                    scenarios, np.int64) and not isinstance(
                        scenarios, str) and not isinstance(
                            scenarios, list):
            raise TypeError(
                "scenarios must be Scenario instance, String, Integer, or List")
        if project is not None and not isinstance(
                project, ps.Project) and not isinstance(
                    project, int) and not isinstance(
                    project, np.int64) and not isinstance(project, str):
            raise TypeError(
                "project must be Project instance, String, or Integer")
        if not isinstance(jobs, int) and not isinstance(jobs, np.int64):
            raise TypeError("jobs must be an Integer")
        if not isinstance(copy_external_inputs, bool):
            raise TypeError("copy_external_inputs must be a Logical")
            
    
    def __initialize_export_args(self, scope, ids, empty, include_key, show_full_paths):
    
        args = ["--export", "--lib=%s" % self.__location]
        
        if scope == "Project" and len(ids) > 0:
            args += ["--pid=%d" % ids]
        
        if scope == "Scenario" and len(ids) > 0:               
            args += ["--sid=%d" % ids]
            
        if empty:
            args += ["--schemaonly"]
            
        if include_key:
            args += ["--includepk"]
        
        if show_full_paths:
            args += ["--extfilepaths"]
            
        return args
    
    def __find_filter_column_args(self, filter_column, filter_value, name, scope,
                                ids):
        
        # Check that filter_value is not None
        if filter_value is None:
            raise ValueError("Must specify a filter_value to filter the " +
                             "filter_column by")
        
        # Check if filter_column exists in Datasheet
        check_args = ["--list", "--columns",
                      "--lib=%s" % self.location, 
                      "--sheet=%s" % name]
        ds_cols = self.__console_to_csv(check_args)
        
        if filter_column not in ds_cols.Name.values:
            raise ValueError(
                f"filter column {filter_column} not in Datasheet {name}")
           
        try:
            
            filter_value = int(filter_value)
            
        except ValueError:
                
            # Initialize Datasheets summary
            # TODO: find out why Library and project scoped datasheets not showing up
            self.__init_datasheets(scope=scope, summary=True)
            
            ds_row = self.__datasheets[self.__datasheets.Name == name]
            
            if ds_row["Is Output"].values[0] == "Yes":
                input_sheet_name = ds_cols[
                    ds_cols.Name == filter_column].Formula1.values[0]
                filter_column = "Name" # TODO: Check if this is true for all cases
                
            else:
                input_sheet_name = name
            
            input_datasheet = self.__slow_query_datasheet(input_sheet_name,
                                                          scope, ids)
            
            if (input_datasheet[filter_column] != filter_value).all():
                raise ValueError(f"filter_value {filter_value} does not "+
                                 f"exist in filter_column {filter_column}")
            
            # Get primary key and ID for filter column/value
            primary_key = input_datasheet.columns[0]
            primary_value = input_datasheet[
                input_datasheet[filter_column] == filter_value
                ][primary_key].values[0]
            filter_column = primary_key
            filter_value = primary_value
        
        finally:

            self.__datasheets = None
            
            return filter_column, str(filter_value)
        
    def __find_datasheets_with_all_cols(self, ids):
        
        # Find out if datasheets contain any data
        optional_args = ["--list", "--datasources",
                         "--lib=%s" % self.location,
                         "--sid=%d" % ids]
        optional_cols = self.__console_to_csv(optional_args)
        optional_cols = optional_cols.replace({"No": False, 
                                               "Yes": True})
        
        if optional_cols["Data Inherited"].sum() > 0:
            add_cols = ["Name", "Data", "Data Inherited",
                        "Data Source"]
        else:
            add_cols = ["Name", "Data"]
        
        # Only retain relevant optional columns
        optional_cols = optional_cols[add_cols]
        
        # Merge optional columns with datasheets and return
        optional_ds = self.__datasheets.merge(optional_cols)
        
        return optional_ds
    
    def __return_summarized_datasheets(self, scope, summary, optional, ids):
    
        self.__init_datasheets(scope=scope, summary=summary)
        
        if optional is False:
            
            return self.__datasheets.iloc[:, 1:4]
        
        elif scope != "Scenario":
            
            return self.__datasheets
        
        else:
            
            optional_ds = self.__find_datasheets_with_all_cols(ids)
    
            return optional_ds
    
    def __return_list_of_full_datasheets(self, scope, args, return_hidden,
                                         ids):

        self.__init_datasheets(scope=scope, summary=True)
        d_summary = self.__datasheets.copy()
                        
        ds_list = []
        
        
        for ds in d_summary["Name"]:
            
            ds = self.__check_datasheet_name(ds)

            # Core LNG package datasheet cannot be exported
            if ds == "core_LNGPackage":
                continue
            
            if return_hidden:
                ds_full = self.__slow_query_datasheet(ds, scope, ids)
            else:
                ds_full = self.__fast_query_datasheet(ds, scope, args)
                
            ds_list.append(ds_full)
            
        return ds_list
    
    def __fast_query_datasheet(self, name, scope, args):
        
        # Refresh datasheets
        self.__datasheets=None
        
        # Add arguments
        fast_query_args = list(args)
        fast_query_args.append("--sheet=%s" % name)
        
        if name.startswith("core"):
            fast_query_args += ["--includesys"]
            
        self.__init_datasheets(scope=scope, summary=False,
                               name=name, args=fast_query_args)
        
        return self.__datasheets
    
    def __slow_query_datasheet(self, input_sheet_name, scope, ids):
        
        tempfile_path = self.__generate_tempfile_path()
        
        try:
            
            # Get datasheet
            args = ["--export", "--lib=%s" % self.location,
                    "--sheet=%s" % input_sheet_name, 
                    "--file=%s" % tempfile_path, "--includepk",
                    "--valsheets", "--extfilepaths", "--force"]
            
            if input_sheet_name.startswith("core"):
                args += ["--includesys"]
            
            # Check the scope of the input_sheet_name
            self.__init_datasheets(scope=scope, summary=True)
            args = self.__find_scope_id_args(input_sheet_name, scope, ids,
                                             args)
            
            self.session._Session__call_console(args)
            ds = pd.read_csv(tempfile_path)
            # ds = self.__remove_unnecessary_datasheet_columns(ds,
            #                                                  input_sheet_name)
        
        finally:
            
            if tempfile_path is not None and os.path.exists(tempfile_path):
                os.remove(tempfile_path)
            
        return ds

    def __delete_datasheet(self, scope, name, ids):

        args = ["--delete", "--data", "--lib=%s" % self.location,
                "--sheet=%s" % name, "--force"]
        if scope == "Project":
            args += ["--pid=%d" % ids]
        if scope == "Scenario":
            args += ["--sid=%d" % ids]
        
        result = self.__session._Session__call_console(args)

        if result.returncode == 0:
            print(f"{name} successfully deleted")
        else:
            raise RuntimeError(result.stderr)

    def __validate_delete_datasheet(self, force, append, scope, data):

        delete_or_warn = None

        if (data.empty) & (scope == "Scenario"):
            delete_or_warn = "delete"
        elif (data.empty) & (force is True):
            delete_or_warn = "delete"
        elif (append is False) & (force is True):
            delete_or_warn = "delete"

        if (force is False) & (append is False) & (scope == "Project"):
            delete_or_warn = "warn"
        elif (force is False) & (data.empty) & (scope == "Project"):
            delete_or_warn = "warn"
        elif (force is False) & (data.empty) & (scope == "Library"):
            delete_or_warn = "warn"

        return delete_or_warn

    def __save_datasheet_to_temp(self, data):

        temp_folder = tempfile.mkdtemp(prefix="SyncroSim-")
        fpath = '{}\\export.csv'.format(temp_folder)
        data.to_csv(fpath, index=False)

        if not os.path.isfile(fpath):
            raise RuntimeError(f"file path {fpath} does not exist")

        return fpath
    
    def __generate_tempfile_path(self):
        
        temp_folder = self.location + "temp"

        if not os.path.exists(temp_folder):
            os.mkdir(temp_folder)
        
        tempfile_path = os.path.join(temp_folder,
                                     "temp.csv")
        
        return tempfile_path
    
    def __find_scope_id_args(self, input_sheet_name, scope, ids, args):
        
        scope_list = ["Project", "Scenario", "Library"]

        if (self.__datasheets.Name == input_sheet_name).any():
            
            input_scope = scope
            
            if input_scope == "Project" and len(ids) > 0:
                args += ["--pid=%d" % ids[0]]
            
            if input_scope == "Scenario" and len(ids) > 0:               
                args += ["--sid=%d" % ids[0]]
                
        else:
            scope_list.remove(scope)
            input_scope = scope_list[0]
                
            if input_scope == "Project" and len(ids) > 0:
                if scope == "Scenario":
                    pid = self.scenarios(sid=ids[0]).project.pid
                else:
                    pid = ids[0]
                args += ["--pid=%d" % pid]
            
            if input_scope == "Scenario" and len(ids) > 0:               
                args += ["--sid=%d" % ids[0]]
                
        return args
    
    def __remove_unnecessary_datasheet_columns(self, ds, input_sheet_name):
        
        ds = ds.dropna(axis=1)
        id_column = input_sheet_name.split("_")[1] + "ID"
        ds = ds.drop([id_column], axis=1)
        
        return ds
    
    def __generate_scenarios_list_to_run(self, scenarios, project):
    
        if scenarios is None:
            
            scenario_list = self.__find_scenarios_from_project(project)
            
            return scenario_list
                    
        elif scenarios is not None:
                
            if not isinstance(scenarios, list):
                scenarios = [scenarios]
                
            if isinstance(scenarios[0], int):
                scenario_list = [
                    self.scenarios(sid=scn) for scn in scenarios]
            elif isinstance(scenarios[0], str):
                scenario_list = [
                    self.scenarios(
                        name=scn, project=project) for scn in scenarios]
            else:
                scenario_list = scenarios
    
            return scenario_list
        
    def __find_scenarios_from_project(self, project):
        
        if project is None:
            
            if len(self.projects()) == 1:
                project = self.projects(
                    pid=self.__projects.ProjectID.item())
                
            else:
                raise ValueError(
                    "Must specify project when > 1 Project in the Library")
        
        # Convert project to Project instance
        if isinstance(project, int):
            project = self.projects(pid=project)
            
        if isinstance(project, str):
            project = self.projects(name=project)
            
        # Run all Scenarios in a Project
        scenario_list = project.scenarios(summary=False)
        
        if not isinstance(scenario_list, list):
            scenario_list = [scenario_list]
            
        return scenario_list
    
    def __get_library_structure(self):

        args = ["--list", "--library", "--lib=%s" % self.location,
                "--tree"]
        lib_structure = self.session._Session__call_console(args, decode=True)
        lib_structure = lib_structure.replace("|", " ")
        lib_structure = lib_structure.split("\r\n")
        lib_structure.remove('')

        level = []
        item = []
        ssim_id = []

        for i in range(0, len(lib_structure)):
            if i == 0:
                level.append(i)
                item.append("Library")
                ssim_id.append(0)
            else:
                l = (len(lib_structure[i]) - len(lib_structure[i].lstrip())) / 3

                start = "+- "
                end = " ["
                it = lib_structure[i][lib_structure[i].find(start) + len(start):lib_structure[i].find(end)]
                it = it.replace("*", "")

                start = "["
                end = "]"
                obj_id = lib_structure[i][lib_structure[i].find(start) + len(start):lib_structure[i].find(end)]

                level.append(l)
                item.append(it)
                ssim_id.append(obj_id)

        return pd.DataFrame({"level": level, "item": item, "id": ssim_id})